--
-- XC9500 CPLD design which controls the configuration of the XSV Virtex
-- with data from the Flash chip.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity config is
	generic
	(
		ADDR_LEN: positive := 21		-- number of Flash address bits
	);
	port
	(
		clk		: in std_logic;	-- clock from DS1075 prog. osc.

		-- Flash address and control pins
		a		: out std_logic_vector(ADDR_LEN-1 downto 0); -- Flash address
		ceb		: out std_logic;	-- Flash chip-enable
		oeb		: out std_logic;	-- Flash output-enable
		web		: out std_logic;	-- Flash write-enable
		resetb		: out std_logic;	-- Flash reset

		-- Virtex configuration pins
		V_progb	: out std_logic;	-- Virtex PROGRAM pin
		V_cclk		: out std_logic;	-- Virtex config clock
		V_csb		: out std_logic;	-- Virtex config chip-select
		V_wrb		: out std_logic;	-- Virtex config write-enable
		V_initb	: in std_logic;	-- Virtex config init status
		V_dout		: in std_logic;	-- Virtex config busy status
		V_done		: in std_logic;	-- Virtex config done status
		V_m		: out std_logic_vector(2 downto 0)	-- Virtex config. mode pins	
	);
end config;

architecture config_arch of config is
	constant LO		: std_logic := '0';
	constant HI		: std_logic := '1';
	constant FLOAT	: std_logic := 'Z';

	signal clk_cnt		: std_logic_vector(3 downto 0);
	signal cclk			: std_logic;
	signal programb, cs		: std_logic;
	signal addr, next_addr	: std_logic_vector(ADDR_LEN-1 downto 0);
	signal poweron_reset	: std_logic;
	signal poweron_cnt		: std_logic_vector(19 downto 0);
	signal V_busy			: std_logic;
	signal button_progb		: std_logic;
begin

	V_busy <= V_dout;		-- give this signal a better name

	-- set Virtex mode to SelectMAP so it can be configured from Flash
	V_m		<= "110";

	-- Flash is enabled for reading while Virtex is not yet configured
	-- and then the Flash pins float when configuration is done
	oeb		<= LO when (V_done=LO) else FLOAT;
	ceb		<= LO when (V_done=LO) else FLOAT;
	web		<= HI when (V_done=LO) else FLOAT;	-- disable Flash writes
	resetb		<= HI;						-- remove Flash reset

	-- generate configuration clock for Virtex from the XSV clock.
	-- The XSV clock could be as much as 100 MHz, so divide by 16
	-- to exceed the access time of the Flash.
	process(clk,clk_cnt)
	begin
		if(clk'event and clk=HI) then
			clk_cnt <= clk_cnt + 1;
		end if;
	end process;
	cclk	<= clk_cnt(3);	-- internal configuration clock
	V_cclk	<= cclk;		-- also send config. clock to Virtex

	-- Apply reset when the power to the XSV Board is first applied.
	-- Remove the power-on reset after the counter reaches 0.
	process(poweron_cnt,cclk)
	begin
		if(cclk'event and cclk=HI) then
			if(poweron_cnt = 0) then
				poweron_reset <= LO;	-- remove reset when timeout expires
			else
				poweron_cnt <= poweron_cnt - 1;
				poweron_reset <= HI;
			end if;
		end if;
	end process;

	-- initiate Virtex configuration by lowering the /PROGRAM pin
	-- during the initial power-on reset and then raising it when
	-- the power-on timeout expires and the manual program control is high
	programb <= not(poweron_reset);
	V_progb <= programb;

	-- Select the Virtex for configuration as long as the /PROGRAM pin
	-- is not held low and the INIT pin is not low.
	process(V_initb,cclk,programb)
	begin
		if(programb = LO) then
			cs <= LO;
		elsif(cclk'event and cclk=HI) then
			cs <= V_initb;
		end if;
	end process;

	-- Select the Virtex for configuration by lowering its chip-select
	-- and write inputs when the internal chip-select is high.  Then
	-- float these pins after the Virtex configuration is done.
	V_csb <= not(cs)	when (V_done=LO) else FLOAT;
	V_wrb <= not(cs)	when (V_done=LO) else FLOAT;

	-- increment the Flash address so the next byte of configuration
	-- data is presented to the Virtex.  Stop incrementing if the
	-- Virtex is not selected, signals a config. error (INIT=0), or
	-- is busy.  Reset the address counter to zero whenever the
	-- /PROGRAM pin goes low and a new configuration sequence begins.
	process(addr,cs,V_initb,V_busy,cclk)
	begin
		if(cclk'event and cclk=HI) then
			if((cs=HI) and (V_initb=HI) and (V_busy=LO)) then
				addr <= addr + 1;
			elsif(programb = LO) then
				addr <= (others=>LO);
			end if;
		end if;
	end process;

	-- pass the Flash address out to the Flash chip.  Float the address
	-- lines once configuration is done.
	a <= addr			when (V_done=LO) else (others=>FLOAT);

end config_arch;

